#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.adapter.file import File
from cliboa.adapter.sftp import SftpAdapter
from cliboa.scenario.sftp import BaseSftp
from cliboa.util.base import _warn_deprecated, _warn_deprecated_args
from cliboa.util.constant import StepStatus


class SftpBaseLoad(BaseSftp):
    def __init__(self):
        super().__init__()
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.load.sftp.SftpBaseLoad",
                "3.0",
                "4.0",
                "cliboa.scenario.sftp.BaseSftp",
            )
        )


class SftpUpload(BaseSftp):
    """
    Upload file to sftp server
    """

    class Arguments(BaseSftp.Arguments):
        dest_dir: str
        quit: bool = False
        ignore_empty_file: bool = False
        put_intermediation: str = "."

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _quit(self):
        return self.args.quit

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _ignore_empty_file(self):
        return self.args.ignore_empty_file

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _put_intermediation(self):
        return self.args.put_intermediation

    def execute(self, *args):
        files = self._resolve("adapter_file", File).get_target_files(
            self.args.src_dir, self.args.src_pattern
        )
        adaptor = self.get_adapter()
        if len(files) > 0:
            for file in files:
                if self.args.ignore_empty_file and os.path.getsize(file) == 0:
                    self.logger.info("0 byte file will no be uploaded %s." % file)
                    continue

                obj = SftpAdapter(
                    host=self.args.host,
                    user=self.args.user,
                    password=self.args.password,
                    key=self.args.key,
                ).put_file(
                    src=file,
                    dest=os.path.join(self.args.dest_dir, os.path.basename(file)),
                    put_intermediation=self.args.put_intermediation,
                    endfile_suffix=self.args.endfile_suffix,
                )
                adaptor.execute(obj)
                self.logger.info("%s is successfully uploaded." % file)
        else:
            self.logger.info(
                "Files to upload do not exist. File pattern: {}".format(
                    os.path.join(self.args.src_dir, self.args.src_pattern)
                )
            )
            if self.args.quit is True:
                return StepStatus.SUCCESSFUL_TERMINATION



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.util.base import _warn_deprecated

global _PROCESS_STORE_CACHE
_PROCESS_STORE_CACHE = {}


class ObjectStore:
    """
    Cache any object.
    This cache class is used when same parameter uses from one STEP to the other.
    """

    @staticmethod
    def put(k, v, quiet: bool = False):
        """
        Put value

        Args:
            k (str): Cache key
            v (dict): Cache value
        """
        if not quiet:
            _warn_deprecated("cliboa.util.cache.ObjectStore.put", "3.0", "4.0")
        _PROCESS_STORE_CACHE[k] = v

    @staticmethod
    def get(k):
        """
        Get value

        Args:
            k (str): Cache key

        Returns:
            dict: Value. Returns None if the key does not exist
        """
        _warn_deprecated("cliboa.util.cache.ObjectStore.get", "3.0", "4.0")
        return _PROCESS_STORE_CACHE.get(k)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
import os
from abc import abstractmethod
from collections import OrderedDict

import yaml

from cliboa.util.base import _BaseObject
from cliboa.util.exception import FileNotFound, ScenarioFileInvalid


class _ScenarioLoader(_BaseObject):
    def __init__(self, scenario_file: str, is_required: bool = False, **kwargs):
        super().__init__(**kwargs)
        self._scenario_file = scenario_file
        if is_required and not self._exists():
            raise FileNotFound("File %s does not exist" % self._scenario_file)

    def _exists(self) -> bool:
        return os.path.isfile(self._scenario_file)

    def __call__(self) -> dict | None:
        """
        Load scenario file and return dict or None.
        """
        if not self._exists():
            return None
        top_dict = self._load()
        if not isinstance(top_dict, dict):
            raise ScenarioFileInvalid(
                f"scenario file {self._scenario_file} is invalid. Check file format."
            )
        return top_dict

    @abstractmethod
    def _load(self) -> dict:
        raise NotImplementedError()


class _YamlScenarioLoader(_ScenarioLoader):
    def _load(self) -> dict:
        with open(self._scenario_file, "r") as f:
            return yaml.safe_load(f)


class _JsonScenarioLoader(_ScenarioLoader):
    def _load(self) -> dict:
        with open(self._scenario_file, "r") as f:
            return json.load(f, object_pairs_hook=OrderedDict)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

from pydantic import BaseModel

from cliboa.adapter.sftp import SftpAdapter
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated, _warn_deprecated_args


class BaseSftp(BaseStep):
    class Arguments(BaseModel):
        src_dir: str
        src_pattern: str
        dest_dir: str = ""
        host: str
        port: int = 22
        user: str
        password: str | None = None
        key: str | None = None
        passphrase: str | None = None
        endfile_suffix: str | None = None
        timeout: int = 30
        retry_count: int = 3

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _src_dir(self):
        return self.args.src_dir

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _src_pattern(self):
        return self.args.src_pattern

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dest_dir(self):
        return self.args.dest_dir

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _host(self):
        return self.args.host

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _port(self):
        return self.args.port

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _user(self):
        return self.args.user

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _password(self):
        return self.args.password

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _key(self):
        return self.args.key

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _passphrase(self):
        return self.args.passphrase

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _endfile_suffix(self):
        return self.args.endfile_suffix

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _timeout(self):
        return self.args.timeout

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _retry_count(self):
        return self.args.retry_count

    def get_adapter(self):
        return self._resolve(
            "adapter_sftp",
            SftpAdapter,
            host=self.args.host,
            user=self.args.user,
            password=self.args.password,
            key=self.args.key,
            passphrase=self.args.passphrase,
            timeout=self.args.timeout,
            retryTimes=self.args.retry_count,
            port=self.args.port,
        )

    def get_adaptor(self):
        self.logger.info(
            _warn_deprecated(
                "cliboa.scenario.sftp.BaseSftp.get_adaptor",
                "3.0",
                "4.0",
                "cliboa.scenario.sftp.BaseSftp.get_adapter",
            )
        )
        return self.get_adapter()



class _StateManager:
    def __init__(self):
        self._state: str = "_Initialized"
        self._steps_max: int = 0
        self._steps_current: int = 0
        self._in_steps: bool = False

    @property
    def current(self) -> str:
        if self._in_steps:
            if self._steps_max >= 10:
                return f"[{self._steps_current:02d}/{self._steps_max:02d}]{self._state}"
            else:
                return f"[{self._steps_current}/{self._steps_max}]{self._state}"
        else:
            return self._state

    @property
    def steps_max(self) -> int:
        return self._steps_max

    def __str__(self) -> str:
        return self.current

    def set(self, new_state: str):
        if not isinstance(new_state, str):
            raise TypeError("The state must be a string.")
        self._state = new_state

    def set_steps_max(self, value: int):
        self._steps_max = value

    def set_steps_current(self, value: int):
        self._steps_current = value

    def set_in_steps(self, value: bool):
        self._in_steps = value


state = _StateManager()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import Field

from cliboa.adapter.csv import Csv
from cliboa.scenario.file import FileRead
from cliboa.scenario.sqlite import SqliteTransaction
from cliboa.util.base import _warn_deprecated_args
from cliboa.util.exception import CliboaException, FileNotFound


class SqliteImport(SqliteTransaction, FileRead):
    class Arguments(SqliteTransaction.Arguments, FileRead.Arguments):
        tblname: str
        primary_key: str | None = None
        index: list[str] = Field(default_factory=list)
        refresh: bool = True
        force_insert: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _tblname(self):
        return self.args.tblname

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _primary_key(self):
        return self.args.primary_key

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _index(self):
        return self.args.index

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _refresh(self):
        return self.args.refresh

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _force_insert(self):
        return self.args.force_insert

    def process(self):
        files = self.get_src_files()
        if not self.check_file_existence(files):
            raise FileNotFound("No csv file was found.")

        files.sort()
        # Find csv columns from all csv files
        csv_columns = []
        for file in files:
            csv_columns.extend(Csv.get_column_names(file))
        csv_columns = sorted(set(csv_columns), key=csv_columns.index)

        if self.args.refresh is True:
            # Drop table in advance, If refresh is True
            self._sqlite_adptr.drop_table(self.args.tblname)
            self._sqlite_adptr.create_table(self.args.tblname, csv_columns, self.args.primary_key)
        else:
            self._sqlite_adptr.create_table(self.args.tblname, csv_columns, self.args.primary_key)

            if self.args.force_insert is True:
                db_columns = self._sqlite_adptr.get_column_names(self.args.tblname)
                result = list(set(csv_columns) - set(db_columns))
                self._sqlite_adptr.add_columns(self.args.tblname, result)
            else:
                # Make sure if csv columns and db table names are exactly the same
                db_columns = self._sqlite_adptr.get_column_names(self.args.tblname)
                if self._sqlite_adptr.escape_columns(
                    csv_columns
                ) != self._sqlite_adptr.escape_columns(db_columns):
                    raise CliboaException(
                        "Csv columns %s were not matched to table column %s."
                        % (csv_columns, db_columns)
                    )

        for file in files:
            self._sqlite_adptr.import_table(
                file, self.args.tblname, refresh=False, encoding=self.args.encoding
            )

        if self.args.index and len(self.args.index) > 0:
            """
            Create index (Add the index at the end for
            better performance when insert data is large)
            """
            self._sqlite_adptr.add_index(self.args.tblname, self.args.index)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
import os
import stat
from time import sleep

from paramiko import AutoAddPolicy, SSHClient

from cliboa.util.base import _BaseObject
from cliboa.util.log import _get_logger


class SftpAdapter(_BaseObject):
    """
    Sftp Adaptor
    """

    TIMEOUT_SEC = 30

    def __init__(
        self,
        host,
        user,
        password=None,
        key=None,
        passphrase=None,
        timeout=TIMEOUT_SEC,
        retryTimes=3,
        port=22,
        **kwargs,
    ):
        """
        Must set whether password or key

        Args:
            host (str): hostname
            user (str): username
            password (str): password
            key (path): private key path
            passphrase (str): passphrase
            timeout=30 (int): timeout seconds
            retryTimes=3 (int): retry count
            port=22 (int): port number
        Raises:
            ValueError: both password and private key are empty, or both specified.
        """
        super().__init__(**kwargs)
        logging.getLogger("paramiko").setLevel(logging.CRITICAL)

        if (not password and not key) or (password and key):
            raise ValueError("Illegal paramters are given.")

        self._host = host
        self._user = user
        self._password = password
        self._key = key
        self._passphrase = passphrase
        self._timeout = timeout
        self._retryTimes = retryTimes
        self._port = 22 if port is None else port

    def execute(self, obj):
        """
        Execute sftp process.
        Pass a result of cliboa.adapter.sftp.Sftp.{{ function }} for the arguments.

        Args:
            obj (tuple):
                obj[0]: A function to be executed.
                obj[1]: Arguments when function is executed.
        """
        return self._execute(obj[0], obj[1])

    def _execute(self, func, kwargs):
        if not func:
            raise ValueError("Function must not be empty.")

        err = None
        for _ in range(self._retryTimes):
            ssh = None
            sftp = None

            try:
                ssh = SSHClient()
                ssh.set_missing_host_key_policy(AutoAddPolicy())  # nosec
                if self._password:
                    ssh.connect(
                        hostname=self._host,
                        username=self._user,
                        password=self._password,
                        port=self._port,
                        timeout=self._timeout,
                    )
                else:
                    ssh.connect(
                        hostname=self._host,
                        username=self._user,
                        key_filename=self._key,
                        passphrase=self._passphrase,
                        port=self._port,
                        timeout=self._timeout,
                    )

                sftp = ssh.open_sftp()

                # call the argument function
                ret = func(sftp=sftp, **kwargs)

                return ret

            except Exception as e:
                self._logger.warning(e)
                self._logger.warning(kwargs)
                err = e

            finally:
                if sftp is not None:
                    sftp.close()
                if ssh is not None:
                    ssh.close()

            self._logger.warning("Unexpected error occurred. Retry will start in 10 sec.")
            sleep(10)

        raise IOError(err, "SFTP failed.")

    def list_files(self, dir, dest, pattern, endfile_suffix=None, ignore_empty_file=False):
        """
        Fetch all the files in specified directory

        Args:
            dir (str): fetch target directory
            dest (str): local directory to save files
            pattern (object): fetch file pattern
            endfile_suffix=None (str): Download a file only if "filename + endfile_suffix" is exists
            ignore_empty_file=False (bool): If True, size zero files are not be downloaded

        Returns (tuple):
            func: list_file_func
            params: parameters for list_file_func

        Raises:
            IOError: failed to get data
        """
        return (
            list_file_func,
            {
                "dir": dir,
                "dest": dest,
                "pattern": pattern,
                "endfile_suffix": endfile_suffix,
                "ignore_empty_file": ignore_empty_file,
            },
        )

    def clear_files(self, dir, pattern):
        """
        Remove all the files which matche to pattern. Not remove directory.

        Args:
            dir (str): remove target dir
            pattern (object): remove target file pattern

        Returns (tuple):
            func: clear_file_func
            params: parameters for clear_file_func

        Raises:
            IOError: failed to remove
        """
        return (clear_file_func, {"dir": dir, "pattern": pattern})

    def remove_specific_file(self, dir, fname):
        """
        Remove specific file (by name) from target directory

        Args:
            dir (str): remove target dir
            fname (str): file name (exact match)

        Returns (tuple):
            func: remove_specific_file_func
            params: parameters for remove_specific_file_func

        Raises:
            IOError: failed to remove
        """
        return (remove_specific_file_func, {"dir": dir, "fname": fname})

    def get_specific_file(self, src, dest):
        """
        Fetch file by exact pattern match

        Args:
            src (str): fetch target absolute path
            dest (str): local directory to save file

        Returns (tuple):
            func: get_specific_file_func
            params: parameters for get_specific_file_func

        Raises:
            IOError: failed to get data
        """
        return (get_specific_file_func, {"src": src, "dest": dest})

    def put_file(self, src, dest, put_intermediation, endfile_suffix=None):
        """
        Upload file to sftp server

        Args:
            src (str): local file to upload
            dest (str): destination sftp path to upload
            endfile_suffix=None (str): Places file with original file name
                            + "endfile_suffix" when upload completed
        Args:
            src (str): fetch target absolute path
            dest (str): local directory to save file

        Returns (tuple):
            func: put_file_func
            params: parameters for put_file_func

        Raises:
            IOError: failed to upload
            :param put_intermediation:
        """
        return (
            put_file_func,
            {
                "src": src,
                "dest": dest,
                "put_intermediation": put_intermediation,
                "endfile_suffix": endfile_suffix,
            },
        )

    def file_exists_check(self, dir, pattern, ignore_empty_file=False):
        """
        Fetch all the files name in specified directory

        Args:
            dir (str): fetch target directory
            pattern (object): fetch file pattern
            ignore_empty_file=False (bool):
             If True, It is treated as if there is no file with size 0.

        Returns (tuple):
            func: file_exists_check_func
            params: parameters for file_exists_check_func

        Raises:
            IOError: failed to get data
        """
        return (
            file_exists_check_func,
            {"dir": dir, "pattern": pattern, "ignore_empty_file": ignore_empty_file},
        )


def list_file_func(**kwargs):
    """
    Get all the files which matches to pattern
    """
    endfile_suffix = kwargs["endfile_suffix"]
    targets = kwargs["sftp"].listdir(kwargs["dir"])
    files = []
    for f in targets:
        if kwargs["pattern"].fullmatch(f) is None:
            continue
        if endfile_suffix and not f + endfile_suffix in targets:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            if kwargs["ignore_empty_file"] is True:
                size = _get_file_size(kwargs["sftp"], fpath)
                if size == 0:
                    continue
            kwargs["sftp"].get(fpath, os.path.join(kwargs["dest"], f))
            files.append(f)
    return files


def clear_file_func(**kwargs):
    """
    All the files which match to pattern.
    """
    files = []
    for f in kwargs["sftp"].listdir(kwargs["dir"]):
        if kwargs["pattern"].fullmatch(f) is None:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            kwargs["sftp"].remove(fpath)
            files.append(fpath)
    return files


def remove_specific_file_func(**kwargs):
    """
    Remove specific file (by name) from target directory
    """
    kwargs["sftp"].remove(os.path.join(kwargs["dir"], kwargs["fname"]))


def get_specific_file_func(**kwargs):
    """
    Get only one files which matches to pattern
    """
    kwargs["sftp"].get(kwargs["src"], kwargs["dest"])


def put_file_func(**kwargs):
    """
    Function that is to upload a file.
    '.' as prefix is added to the file name
    while uploading and then the prefix '.'
    will be removed (rename the file) when upload complete.
    The function to add and remove the prefix can be disabled
    by specifying "None" for "put_intermediation".
    """
    dirname = os.path.dirname(kwargs["dest"])

    # Create directory if it doesn't exist
    try:
        kwargs["sftp"].stat(dirname)
    except FileNotFoundError:
        kwargs["sftp"].mkdir(dirname)

    # Same file name is removed in advance, if exists
    try:
        f = kwargs["sftp"].stat(kwargs["dest"])
        if f:
            kwargs["sftp"].remove(kwargs["dest"])
    except FileNotFoundError:
        pass

    if kwargs["put_intermediation"] is None:
        kwargs["sftp"].put(kwargs["src"], os.path.join(dirname, os.path.basename(kwargs["dest"])))
    else:
        tmp_dest = os.path.join(
            dirname, kwargs["put_intermediation"] + os.path.basename(kwargs["dest"])
        )
        kwargs["sftp"].put(kwargs["src"], tmp_dest)
        kwargs["sftp"].rename(tmp_dest, kwargs["dest"])

    endfile_suffix = kwargs["endfile_suffix"]
    if endfile_suffix is not None:
        endfile = kwargs["src"] + endfile_suffix
        open(endfile, mode="w").close()
        kwargs["sftp"].put(endfile, kwargs["dest"] + endfile_suffix)
        os.remove(endfile)


def file_exists_check_func(**kwargs):
    """
    Get all the file names which matches to pattern
    """
    files = []
    for f in kwargs["sftp"].listdir(kwargs["dir"]):
        if kwargs["pattern"].fullmatch(f) is None:
            continue

        fpath = os.path.join(kwargs["dir"], f)
        if _is_file(kwargs["sftp"], fpath):
            if kwargs["ignore_empty_file"] is True:
                if _get_file_size(kwargs["sftp"], fpath) == 0:
                    continue
            files.append(f)
    return files


def _transfer_with_callback(reader, writer, file_size, callback):
    """
    For Debug
    """
    size = 0
    _logger = _get_logger(__name__)
    _logger.debug("Call back method")
    while True:
        data = reader.read(32768)
        writer.write(data)
        size += len(data)
        if len(data) == 0:
            _logger.debug("No data")
            break
        if callback is not None:
            callback(size, file_size)
    _logger.debug("Size %s" % size)
    return size


def _is_file(sftp, path):
    """
    Returns True if target object is a file, False if not.
    """
    info = sftp.stat(path)
    if stat.S_ISREG(info.st_mode):
        return True
    else:
        return False


def _get_file_size(sftp, path):
    """
    Returns a file size
    """
    info = sftp.stat(path)
    return info.st_size



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import json
import os
import re
from datetime import datetime

from cliboa.adapter.gcp import BigQueryAdapter, FireStoreAdapter, GcsAdapter
from cliboa.scenario.gcp import BaseBigQuery, BaseFirestore, BaseGcs
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.constant import StepStatus
from cliboa.util.exception import InvalidParameter
from cliboa.util.string import StringUtil


class BigQueryRead(BaseBigQuery):
    """
    Read data from BigQuery and put them into on-memory or export to a file via GCS.
    """

    _RANDOM_STR_LENGTH = 8

    def __init__(self):
        super().__init__()
        self._key = None
        self._bucket = None
        self._dest_dir = None
        self._filename = None
        self._query = None

    def key(self, key):
        self._key = key

    def query(self, query):
        self._query = query

    def bucket(self, bucket):
        self._bucket = bucket

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def filename(self, filename):
        self._filename = filename

    def execute(self, *args):
        super().execute()
        if not self._key and not self._bucket:
            raise InvalidParameter("Specifying either 'key' or 'bucket' is essential.")
        if self._key and self._bucket:
            raise InvalidParameter("Cannot specify both 'key' and 'bucket'.")

        # fetch records and save to on-memory
        if self._key:
            valid = EssentialParameters(self.__class__.__name__, [self._tblname])
            valid()
            self._save_to_cache()
        elif self._bucket:
            valid = EssentialParameters(self.__class__.__name__, [self._dest_dir])
            valid()
            self._save_as_file_via_gcs()

    def _save_to_cache(self):
        self._logger.info("Save data to on memory")

        client = BigQueryAdapter().get_client(
            credentials=self.get_credentials(),
            project=self._project_id,
            location=self._location,
        )

        query = "SELECT * FROM %s.%s" % (self._dataset, self._tblname)
        query = self._query if self._query else query

        df = client.query(query).to_dataframe()
        self.put_to_context({f"{self._key}": df})

    def _save_as_file_via_gcs(self):
        self._logger.info("Save data as a file via GCS")
        os.makedirs(self._dest_dir, exist_ok=True)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        path = "%s-%s" % (
            StringUtil().random_str(self._RANDOM_STR_LENGTH),
            ymd_hms,
        )
        prefix = "%s/%s/%s" % (self._dataset, self._tblname, path)

        gbq_client = BigQueryAdapter().get_client(credentials=self.get_credentials())

        if self._dataset and self._tblname:
            table_ref = gbq_client.dataset(self._dataset).table(self._tblname)
        elif self._dataset and not self._tblname:
            tmp_tbl = "tmp_" + StringUtil().random_str(self._RANDOM_STR_LENGTH) + "_" + ymd_hms
            table_ref = gbq_client.dataset(self._dataset).table(tmp_tbl)

        gcs_client = GcsAdapter().get_client(credentials=self.get_credentials())
        gcs_bucket = gcs_client.bucket(self._bucket)

        # extract job config settings
        ext_job_config = BigQueryAdapter.get_extract_job_config()
        ext_job_config.compression = BigQueryAdapter.get_compression_type()
        ext = ".csv"
        if self._filename:
            _, ext = os.path.splitext(self._filename)
            support_ext = [".csv", ".json"]
            if ext not in support_ext:
                raise InvalidParameter("%s is not supported as filename." % ext)
        ext_job_config.destination_format = BigQueryAdapter.get_destination_format(ext)

        comp_format_and_ext = {"GZIP": ".gz"}
        comp_ext = comp_format_and_ext.get(str(BigQueryAdapter.get_compression_type()))
        if self._filename:
            dest_gcs = "gs://%s/%s/%s%s" % (
                self._bucket,
                prefix,
                self._filename,
                comp_ext,
            )
        else:
            dest_gcs = "gs://%s/%s/*%s%s" % (self._bucket, prefix, ext, comp_ext)

        # Execute query.
        if self._query:
            query_job_config = BigQueryAdapter.get_query_job_config()
            query_job_config.destination = table_ref
            query_job_config.write_disposition = BigQueryAdapter.get_write_disposition()
            query_job = gbq_client.query(
                self._query, location=self._location, job_config=query_job_config
            )
            query_job.result()

        # Extract to GCS
        extract_job = gbq_client.extract_table(
            table_ref, dest_gcs, job_config=ext_job_config, location=self._location
        )
        extract_job.result()

        # Download from gcs
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            dest = os.path.join(self._dest_dir, os.path.basename(blob.name))
            blob.download_to_filename(dest)

        # Cleanup temporary table
        if self._query:
            gbq_client.delete_table(table_ref)

        # Cleanup temporary files
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            blob.delete()


class GcsDownload(BaseGcs):
    """
    Download from GCS
    """

    def __init__(self):
        super().__init__()
        self._prefix = None
        self._delimiter = None
        self._src_pattern = None
        self._dest_dir = "."

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        client = GcsAdapter().get_client(credentials=self.get_credentials())
        bucket = client.bucket(self._bucket)
        dl_files = []
        for blob in client.list_blobs(bucket, prefix=self._prefix, delimiter=self._delimiter):
            r = re.compile(self._src_pattern)
            if not r.fullmatch(os.path.basename(blob.name)):
                continue
            dl_files.append(blob.name)
            blob.download_to_filename(os.path.join(self._dest_dir, os.path.basename(blob.name)))

        self.put_to_context(dl_files)


class GcsDownloadFileDelete(BaseGcs):
    """
    Delete files downloaded from GCS by using cache
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        dl_files = self.get_from_context()

        if len(dl_files) > 0:
            self._logger.info("Delete files %s" % dl_files)
            client = GcsAdapter().get_client(credentials=self.get_credentials())
            bucket = client.bucket(self.get_symbol_argument("bucket"))
            for blob in client.list_blobs(
                bucket,
                prefix=self.get_symbol_argument("prefix"),
                delimiter=self.get_symbol_argument("delimiter"),
            ):
                for dl_f in dl_files:
                    if dl_f == blob.name:
                        blob.delete()
                        break
        else:
            self._logger.info("No files to delete.")


class FirestoreDocumentDownload(BaseFirestore):
    """
    Download a document from Firestore
    """

    def __init__(self):
        super().__init__()
        self._dest_dir = None

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._collection, self._document, self._dest_dir]
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        client = FireStoreAdapter().get_client(self.get_credentials())
        ref = client.document(self._collection, self._document)
        doc = ref.get()

        with open(os.path.join(self._dest_dir, doc.id), mode="wt") as f:
            f.write(json.dumps(doc.to_dict()))


class GcsFileExistsCheck(BaseGcs):
    """
    File check in GCS
    """

    def __init__(self):
        super().__init__()
        self._prefix = None
        self._delimiter = None
        self._src_pattern = None

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        client = GcsAdapter().get_client(credentials=self.get_credentials())
        dl_files = []
        bucket = client.bucket(self.get_symbol_argument("bucket"))

        for blob in client.list_blobs(bucket, prefix=self._prefix, delimiter=self._delimiter):
            r = re.compile(self._src_pattern)
            if not r.fullmatch(os.path.basename(blob.name)):
                continue
            dl_files.append(blob.name)

        if len(dl_files) == 0:
            self._logger.info("File not found in GCS. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        self._logger.info("File was found in GCS. After process will be processed")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import functools
import inspect
import warnings
from abc import ABC
from typing import TypeVar

from cliboa.util.exception import CliboaException
from cliboa.util.log import _get_logger

T = TypeVar("T")


class _BaseObject(ABC):
    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except TypeError as e:
            if "object.__init__()" in str(e):
                super().__init__()
            else:
                raise e
        self._di_map = {}
        self._di_kwargs = {}
        for k, v in kwargs.items():
            if k.startswith("di_"):
                self._di_kwargs[k] = v
                key = k[3:]
                self._di_map[key] = v
        self._logger = self._resolve(
            "logger", _get_logger(self.__class__.__module__ + "." + self.__class__.__name__)
        )

    def _resolve_cls(self, key: str, default_cls: type[T], *args, **kwargs) -> type[T]:
        dependency = self._di_map.get(key, default_cls)
        if not inspect.isclass(dependency):
            raise CliboaException(f"Failed to resolve cliboa class. {key}:{type(dependency)}")
        return dependency

    def _resolve(self, _di_key: str, _default_cls: type[T], *args, **kwargs) -> T:
        dependency = self._di_map.get(_di_key, _default_cls)
        if not inspect.isclass(dependency):
            return dependency
        merged_kwargs = self._di_kwargs | kwargs
        return dependency(*args, **merged_kwargs)


def _warn_deprecated(
    deprecated: str,
    end_version: str,
    removal_version: str,
    instead: str | None = None,
    stacklevel: int = 3,
) -> str:
    err_mes = f"{deprecated} is deprecated."
    if instead:
        err_mes += f" Use {instead} instead."
    err_mes += (
        f" It is no longer supported since version {end_version},"
        f" and it will be removed in version {removal_version}."
    )
    warnings.warn(
        err_mes,
        DeprecationWarning,
        stacklevel=stacklevel,
    )
    return err_mes


def _warn_deprecated_args(end_version: str, removal_version: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            instance = args[0] if args else None
            full_name = f"{func.__module__}.{func.__qualname__}"
            skip_log = False
            if instance and hasattr(instance, "_deprecated_warn_log"):
                if full_name in instance._deprecated_warn_log:
                    skip_log = True
                else:
                    instance._deprecated_warn_log.append(full_name)

            if not skip_log:
                full_split = full_name.split(".")
                prop_name = full_split.pop()
                if prop_name.startswith("_"):
                    prop_name = prop_name[1:]
                full_split.append("args")
                full_split.append(prop_name)
                instead_name = ".".join(full_split)

                log_mes = _warn_deprecated(full_name, end_version, removal_version, instead_name)
                if instance and hasattr(instance, "logger"):
                    instance.logger.warning(log_mes)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _warn_removed(deprecated: str, removal_version: str, stacklevel: int = 3) -> str:
    err_mes = (
        f"{deprecated} has been removed in version {removal_version}."
        " Refer to the documentation for migration details."
    )
    warnings.warn(
        err_mes,
        DeprecationWarning,
        stacklevel=stacklevel,
    )
    return err_mes



import os

from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.gpg import Gpg


class GpgBase(BaseStep):
    def __init__(self):
        super().__init__()
        self._gnupghome = None
        self._passphrase = None
        self._src_dir = None
        self._src_pattern = None
        self._dest_dir = None
        self._always_trust = False
        self._key_dir = None
        self._key_pattern = None
        self._trust_level = None

    def gnupghome(self, gnupghome):
        self._gnupghome = gnupghome

    def passphrase(self, passphrase):
        self._passphrase = passphrase

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def always_trust(self, always_trust):
        self._always_trust = always_trust

    def key_dir(self, key_dir):
        self._key_dir = key_dir

    def key_pattern(self, key_pattern):
        self._key_pattern = key_pattern

    def trust_level(self, trust_level):
        self._trust_level = trust_level

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._gnupghome])
        valid()

    def key_import(self, gpg, key_files, trust_level=None):
        """
        Import key.
        Also set trust level, if parameter "trust_level" is given.
        """
        for file in key_files:
            gpg.import_key(file)
            if trust_level:
                gpg.trust_key(file, trust_level)


class GpgGenerateKey(GpgBase):
    """
    Generate rsa key.
    """

    def __init__(self):
        super().__init__()

        self._name_real = None
        self._name_email = None

    def name_real(self, name_real):
        self._name_real = name_real

    def name_email(self, name_email):
        self._name_email = name_email

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._dest_dir, self._name_email])
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        Gpg(self._gnupghome).generate_key(
            self._dest_dir,
            name_real=self._name_real,
            name_email=self._name_email,
            passphrase=self._passphrase,
        )


class GpgEncrypt(GpgBase):
    """
    Gpg encryption.
    """

    def __init__(self):
        super().__init__()
        self._recipients = []

    def recipients(self, recipients):
        self._recipients = recipients

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._recipients])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            self._logger.info("No files are found. Nothing to do.")
            return

        gpg = Gpg(self._gnupghome)

        if self._key_dir and self._key_pattern:
            key_files = super().get_target_files(self._key_dir, self._key_pattern)
            self._logger.info("Keys found %s" % key_files)
            self.key_import(gpg, key_files, self._trust_level)

        for file in files:
            if self._dest_dir is not None:
                os.makedirs(self._dest_dir, exist_ok=True)
                dest_path = os.path.join(self._dest_dir, os.path.basename(file))
            else:
                dest_path = os.path.join(self._src_dir, os.path.basename(file))
            gpg.encrypt(
                file,
                dest_path,
                recipients=self._recipients,
                passphrase=self._passphrase,
                always_trust=self._always_trust,
            )


class GpgDecrypt(GpgBase):
    """
    Gpg decryption.
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        super().execute()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            self._logger.info("No files are found. Nothing to do.")
            return

        if self._key_dir and self._key_pattern:
            key_files = super().get_target_files(self._key_dir, self._key_pattern)
            self._logger.info("Keys found %s" % key_files)
            self.key_import(key_files, self._trust_level)

        gpg = Gpg(self._gnupghome)
        for file in files:
            root, ext = os.path.splitext(file)
            if ext == ".gpg":
                if self._dest_dir is not None:
                    os.makedirs(self._dest_dir, exist_ok=True)
                    dest_path = os.path.join(self._dest_dir, os.path.basename(root))
                else:
                    dest_path = os.path.join(self._src_dir, os.path.basename(root))
                gpg.decrypt(
                    file,
                    dest_path,
                    passphrase=self._passphrase,
                    always_trust=self._always_trust,
                )
            else:
                self._logger.warning("Extention was not gpg. %s" % file)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import re

from cliboa.adapter.ftp import FtpAdapter
from cliboa.scenario.ftp import BaseFtp
from cliboa.util.base import _warn_deprecated
from cliboa.util.constant import StepStatus


class FtpExtract(BaseFtp):
    def __init__(self):
        super().__init__()
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.extract.ftp.FtpExtract",
                "3.0",
                "4.0",
                "cliboa.scenario.ftp.BaseFtp",
            )
        )


class FtpDownload(BaseFtp):
    """
    Download files from ftp server
    """

    class Arguments(BaseFtp.Arguments):
        dest_dir: str
        quit: bool = False

    def execute(self, *args):
        os.makedirs(self.args.dest_dir, exist_ok=True)

        adapter = self.get_adapter()
        obj = adapter.list_files(
            dir=self.args.src_dir,
            dest=self.args.dest_dir,
            pattern=re.compile(self.args.src_pattern),
        )
        files = adapter.execute(obj)

        if self.args.quit is True and len(files) == 0:
            self.logger.info("No file was found. After process will not be processed")
            return StepStatus.SUCCESSFUL_TERMINATION

        # cache downloaded file names
        self.put_to_context(files)


class FtpDownloadFileDelete(FtpExtract):
    """
    Delete all downloaded files.
    """

    def execute(self, *args):
        files = self.get_from_context()

        if files is None or len(files) == 0:
            self.logger.info("No files to delete.")
            return

        self.logger.info("Delete files %s" % files)

        adapter = self._resolve(
            "adapter_ftp",
            FtpAdapter,
            host=self.get_step_argument("host"),
            user=self.get_step_argument("user"),
            password=self.get_step_argument("password"),
            timeout=self.get_step_argument("timeout"),
            retryTimes=self.get_step_argument("retry_count"),
            port=self.get_step_argument("port"),
            tls=self.get_step_argument("tls"),
        )

        for file in files:
            obj = adapter.remove_specific_file(
                dir=self.get_step_argument("src_dir"),
                fname=file,
            )
            adapter.execute(obj)
            self.logger.info("%s is successfully deleted." % file)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import json
import os

import boto3

from cliboa.adapter.aws import S3Adapter
from cliboa.scenario.aws import BaseAws, BaseS3
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.constant import StepStatus
from cliboa.util.exception import FileNotFound, InvalidFormat, InvalidParameter


class S3Upload(BaseS3):
    """
    Upload to S3
    """

    def __init__(self):
        super().__init__()
        self._key = None
        self._src_dir = None
        self._src_pattern = None
        self._quit = False

    def key(self, key):
        self._key = key

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def quit(self, quit):
        self._quit = quit

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self._src_pattern])
        valid()

        adapter = S3Adapter(
            self._access_key, self._secret_key, self._profile, self._role_arn, self._external_id
        )
        resource = adapter.get_resource()
        bucket = resource.Bucket(self._bucket)
        files = super().get_target_files(self._src_dir, self._src_pattern)

        if len(files) > 0:
            for f in files:
                if self._key:
                    if not self._key.endswith("/"):
                        self._key = self._key + "/"
                    s = "{}{}".format(self._key, os.path.basename(f))
                    bucket.upload_file(Key=s, Filename=f)
                else:
                    bucket.upload_file(Key=os.path.basename(f), Filename=f)
        else:
            self._logger.info(
                "Files to upload do not exist. File pattern: {}".format(
                    os.path.join(self._src_dir, self._src_pattern)
                )
            )
            if self._quit is True:
                return StepStatus.SUCCESSFUL_TERMINATION


class DynamoDBWrite(BaseAws):
    """
    Class to write data from CSV or JSONL files to DynamoDB
    """

    def __init__(self):
        super().__init__()
        self._table_name = None
        self._src_dir = None
        self._src_pattern = None
        self._file_format = "csv"

    def table_name(self, table_name):
        self._table_name = table_name

    def src_dir(self, src_dir):
        self._src_dir = src_dir

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def file_format(self, file_format):
        if file_format not in ["csv", "jsonl"]:
            raise InvalidParameter("file_format must be either 'csv' or 'jsonl'")
        self._file_format = file_format

    def execute(self, *args):
        valid = EssentialParameters(
            self.__class__.__name__,
            [self._table_name, self._src_dir, self._src_pattern, self._file_format],
        )
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) == 0:
            raise FileNotFound("Target file was not found.")

        dynamodb = boto3.resource("dynamodb", region_name=self._region)

        try:
            table = dynamodb.Table(self._table_name)
            table.load()

            table_info = table.key_schema
            partition_key = next(key for key in table_info if key["KeyType"] == "HASH")[
                "AttributeName"
            ]
            sort_key = next(
                (key["AttributeName"] for key in table_info if key["KeyType"] == "RANGE"), None
            )
        except dynamodb.meta.client.exceptions.ResourceNotFoundException:
            self._logger.error(f"Table '{self._table_name}' does not exist.")
            raise

        for file in files:
            if self._file_format.lower() == "csv":
                self._write_csv_to_dynamodb(file, table, partition_key, sort_key)
            elif self._file_format.lower() == "jsonl":
                self._write_jsonl_to_dynamodb(file, table, partition_key, sort_key)
            else:
                raise InvalidParameter(f"Unsupported file type: {self._file_format}")

        self._logger.info("Data has been written to DynamoDB.")

    def _write_csv_to_dynamodb(self, file, table, partition_key, sort_key):
        """Write CSV file to DynamoDB
        Args:
            file (str): File path
            table (boto3.resources.factory.dynamodb.Table): DynamoDB table
            partition_key (str): Partition key
            sort_key (str): Sort key
        """
        with open(file, "r") as f:
            reader = csv.DictReader(f)
            headers = reader.fieldnames
            if partition_key not in headers:
                raise InvalidParameter(
                    f"Partition key '{partition_key}' not found in CSV headers of file: {file}"
                )
            if sort_key and sort_key not in headers:
                raise InvalidParameter(
                    f"Sort key '{sort_key}' not found in CSV headers of file: {file}"
                )

            with table.batch_writer() as batch:
                for row in reader:
                    batch.put_item(Item=row)

    def _write_jsonl_to_dynamodb(self, file, table, partition_key, sort_key):
        """Write JSONL file to DynamoDB
        Args:
            file (str): File path
            table (boto3.resources.factory.dynamodb.Table): DynamoDB table
            partition_key (str): Partition key
            sort_key (str): Sort key
        """
        with open(file, "r") as f:
            with table.batch_writer() as batch:
                for line_number, line in enumerate(f, 1):
                    try:
                        item = json.loads(line)
                    except json.JSONDecodeError as e:
                        raise InvalidFormat(
                            f"Invalid JSON format in file {file} at line {line_number}: {str(e)}"
                        )

                    if partition_key not in item:
                        raise InvalidParameter(
                            f"Partition key '{partition_key}' not found in JSONL file: "
                            f"{file} at line {line_number}"
                        )
                    if sort_key and sort_key not in item:
                        raise InvalidParameter(
                            f"Sort key '{sort_key}' not found in JSONL file: "
                            f"{file} at line {line_number}"
                        )
                    batch.put_item(Item=item)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa import state
from cliboa.core.builder import _ScenarioBuilder
from cliboa.core.executor import _ScenarioExecutor
from cliboa.core.model import CommandArgument
from cliboa.listener.scenario import ScenarioStatusListener
from cliboa.util.base import _BaseObject


class ScenarioManager(_BaseObject):
    """
    Main class to execute pipeline by scenario files.
    """

    def __init__(
        self,
        scenario_file: str,
        common_file: str | list[str] | None = None,
        file_format: str = "yaml",
        cmd_arg: CommandArgument | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._builder = self._resolve(
            "scenario_builder", _ScenarioBuilder, scenario_file, common_file, file_format, cmd_arg
        )

    def execute(self) -> int:
        """
        Main logic of cliboa.
        """
        # 1. Load scenario files and build scenario step instances which are ready to execute.
        state.set("_BuildScenario")
        steps = self._builder.execute()

        # 2. Prepare the steps by wrapping them in an executor instance.
        state.set("_PrepareScenario")
        executor = self._resolve("scenario_executor", _ScenarioExecutor, steps)
        executor.register_listener(
            self._resolve("scenario_status_listener", ScenarioStatusListener)
        )

        # 3. Execute the scenario and return the int result code.
        state.set("_ExecuteScenario")
        return executor.execute()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import subprocess

from pydantic import BaseModel

from cliboa.scenario.base import BaseStep


class _CommandDict(BaseModel):
    content: str | None = None
    file: str | None = None


class ExecuteShellScript(BaseStep):
    """
    Execute Shell Script
    """

    class Arguments(BaseModel):
        command: _CommandDict | str
        work_dir: str | None = None

    def execute(self, *args):
        # Setting Up Directory
        default_dir = os.getcwd()

        try:
            if self.args.work_dir:
                os.chdir(self.args.work_dir)

            if isinstance(self.args.command, str):
                content = self.args.command
            else:
                content = self.args.command.content

            if content:
                for command in content.split("&&"):
                    subprocess.run(command.strip().split(" "))
            elif self.args.command.file:
                subprocess.call(self.args.command.file)
        finally:
            # Set Directory to default
            if self.args.work_dir:
                os.chdir(default_dir)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import json
import os

import pandas

from cliboa.adapter.gcp import BigQueryAdapter, FireStoreAdapter, GcsAdapter, ServiceAccount
from cliboa.scenario.file import FileRead
from cliboa.scenario.gcp import BaseBigQuery, BaseFirestore, BaseGcs
from cliboa.util.exception import FileNotFound, InvalidFormat


class BigQueryWrite(BaseBigQuery, FileRead):
    """
    Read csv and Insert data into BigQuery table
    """

    # default bulk line count to change to dataframe object
    _BULK_LINE_CNT = 10000

    # BigQuery insert mode
    _REPLACE = "replace"
    _APPEND = "append"

    def __init__(self):
        super().__init__()
        self._columns = []

    class Arguments(BaseBigQuery.Arguments, FileRead.Arguments):
        table_schema: list[dict]
        replace: bool = True
        has_header: bool = True

    def execute(self, *args):
        files = self.get_src_files()
        if not self.check_file_existence(files):
            return 1
        self.logger.info("insert target files %s" % files)

        is_inserted = False
        # initial if_exists
        if_exists = self._REPLACE if self.args.replace is True else self._APPEND
        self._columns = [name_and_type["name"] for name_and_type in self.args.table_schema]
        for file in files:
            insert_rows = []
            with open(file, "r", encoding=self.args.encoding) as f:
                reader = (
                    csv.DictReader(f, delimiter=",")
                    if self.args.has_header is True
                    else csv.reader(f, delimiter=",")
                )
                if self.args.has_header is True:
                    for r in reader:
                        # extract only the specified columns
                        contents = {}
                        for c in self._columns:
                            if not r.get(c):
                                continue
                            contents[c] = r.get(c)
                        insert_rows.append(contents)

                        # bulk insert
                        if len(insert_rows) == self._BULK_LINE_CNT:
                            self._exec_insert(insert_rows, is_inserted, if_exists)
                            insert_rows.clear()
                            is_inserted = True
                    if len(insert_rows) > 0:
                        self._exec_insert(insert_rows, is_inserted, if_exists)
                        is_inserted = True

                else:
                    # csv headers do not exist
                    for row in reader:
                        contents = {}
                        for i, c in enumerate(self._columns):
                            contents[c] = row[i]
                        insert_rows.append(contents)

                        # bulk insert
                        if len(insert_rows) == self._BULK_LINE_CNT:
                            self._exec_insert(insert_rows, is_inserted, if_exists)
                            insert_rows.clear()
                            is_inserted = True
                    if len(insert_rows) > 0:
                        self._exec_insert(insert_rows, is_inserted, if_exists)
                        is_inserted = True

    def _exec_insert(self, insert_rows, is_inserted, if_exists):
        """
        Execute insert into a BigQuery table
        Args:
            insert_rows: rows to insert
            is_inserted: if the data is already inserted or not
            if_exists: replace or append
        """
        df = pandas.DataFrame(self._format_insert_data(insert_rows))
        if is_inserted is True:
            # if_exists after the first insert execution
            if_exists = self._APPEND
        dest_tbl = self.args.dataset + "." + self.args.tblname
        self.logger.info("Start insert %s rows to %s" % (len(insert_rows), dest_tbl))

        df.to_gbq(
            dest_tbl,
            project_id=self.args.project_id,
            if_exists=if_exists,
            table_schema=self.args.table_schema,
            location=self.args.location,
            credentials=ServiceAccount().auth(self.get_credentials()),
        )

    def _format_insert_data(self, insert_rows):
        """
        Format insert data to pass DataFrame as the below.

        insert_data = {
            "column1": [1, 2],
            "column2": ["spam", "spam"],
            ...
        }

        Args
            insert_rows: dictionary list of input cache
        """
        insert_data = {}
        for c in self._columns:
            v_list = [d.get(c) for d in insert_rows]
            if not v_list:
                raise InvalidFormat("Specified column %s does not exist in an input file." % c)
            insert_data[c] = v_list
        return insert_data


class GcsUpload(BaseGcs, FileRead):
    """
    Upload local files to GCS
    """

    class Arguments(BaseGcs.Arguments, FileRead.Arguments):
        dest_dir: str = ""

    def execute(self, *args):
        gcs_client = GcsAdapter().get_client(credentials=self.get_credentials())
        bucket = gcs_client.bucket(self.args.bucket)
        files = self.get_src_files()
        self.logger.info("Upload files %s" % files)
        for file in files:
            self.logger.info("Start upload %s" % file)
            blob = bucket.blob(os.path.join(self.args.dest_dir, os.path.basename(file)))
            blob.upload_from_filename(file)
            self.logger.info("Finish upload %s" % file)


class FirestoreDocumentCreate(BaseFirestore, FileRead):
    """
    Create document on FIrestore
    """

    class Arguments(BaseFirestore.Arguments, FileRead.Arguments):
        pass

    def execute(self, *args):
        files = self.get_src_files()
        if not self.check_file_existence(files):
            raise FileNotFound("No files are found.")

        firestore_client = FireStoreAdapter().get_client(self.get_credentials())
        for file in files:
            with open(file) as f:
                fname = os.path.splitext(os.path.basename(file))[0]
                doc = firestore_client.collection(self.args.collection).document(fname)
                doc.set(json.load(f))


class BigQueryCopy(BaseBigQuery):
    """
    Copy Bigquery from one table to another
    """

    class Arguments(BaseBigQuery.Arguments):
        dest_dataset: str
        dest_tblname: str

    def execute(self, *args):
        # Define Source Table and Destination Table
        source_table_id = "{}.{}.{}".format(
            self.args.project_id, self.args.dataset, self.args.tblname
        )
        destination_table_id = "{}.{}.{}".format(
            self.args.project_id, self.args.dest_dataset, self.args.dest_tblname
        )

        # Client Setup
        gbq_client = BigQueryAdapter().get_client(
            credentials=self.get_credentials(),
            project=self.args.project_id,
            location=self.args.location,
        )

        # Create Copy Job
        job = gbq_client.copy_table(source_table_id, destination_table_id)

        # Wait for the job to complete.
        job.result()

        self.logger.info("A copy of the table created.")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.scenario.base import BaseStep


class SampleStep(BaseStep):
    """
    For unit test
    """

    class Arguments(BaseModel):
        retry_count: int = 3
        memo: str | None = None

    def execute(self):
        self.logger.info("Start %s" % self.__class__.__name__)
        self.logger.info(f"my memo is {self.args.memo}")
        self.logger.info("Finish %s" % self.__class__.__name__)


class SampleStepSub(SampleStep):
    """
    For unit test
    """

    class Arguments(SampleStep.Arguments):
        name: str | None = None

    def execute(self, **kwargs):
        self.logger.info(f"Start {self}")
        self.logger.info(f"kwargs is {kwargs}")
        self.logger.info(f"my name is {self.args.name}")
        self.logger.info(f"my memo is {self.args.memo}")
        symbol_memo = self.get_symbol_argument("memo")
        self.logger.info(f"symbol memo is {symbol_memo}")
        symbol_context = self.get_from_context()
        self.logger.info(f"symbol context is {symbol_context}")
        self.logger.info("Finish %s" % self.__class__.__name__)


class SampleCustomStep(BaseStep):
    """
    For unit test
    """

    class Arguments(BaseModel):
        password: str | None = None
        access_key: str | None = None
        secret_key: str | None = None
        access_token: str | None = None
        retry_count: int = 3

    def execute(self):
        pass



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.validator import EssentialParameters as _ValidEP
from cliboa.util.base import _warn_deprecated

_warn_deprecated("cliboa.core.validator", "3.0", "4.0")


class EssentialParameters(_ValidEP):
    """
    DEPRECATED: Use cliboa.scenario.validator.EssentialParameters instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._logger.warning(
            _warn_deprecated(
                ".".join(("cliboa.core.validator", self.__class__.__name__)),
                "2.6.0beta",
                "4.0",
                ".".join(("cliboa.scenario.validator", self.__class__.__name__)),
            )
        )



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import errno
import os
from datetime import datetime
from ftplib import FTP, FTP_TLS  # nosec
from time import sleep

from cliboa.util.base import _BaseObject


class FtpAdapter(_BaseObject):
    """
    Ftp Adaptor
    """

    TIMEOUT_SEC = 30
    RETRY_SEC = 10

    def __init__(
        self, host, user, password, timeout=TIMEOUT_SEC, retryTimes=3, port=21, tls=False, **kwargs
    ):
        """
        Must set whether password or key

        Args:
            host (str): hostname
            user (str): username
            password (str): password
            timeout=30 (int): timeout seconds
            retryTimes=3 (int): retry count
            port=21 (int): port number
            tls=False (bool): use secure connection
        """
        super().__init__(**kwargs)

        self._host = host
        self._user = user
        self._password = password
        self._timeout = timeout
        self._retryTimes = retryTimes
        self._port = port
        self._tls = tls

    def execute(self, obj):
        """
        Execute sftp process.
        Pass a result of cliboa.adapter.ftp.FtpAdapter.{{ function }} for the arguments.

        Args:
            obj (tuple):
                obj[0]: A function to be executed.
                obj[1]: Arguments when function is executed.
        """
        return self._execute(obj[0], obj[1])

    def _execute(self, func, kwargs):

        if not func:
            raise ValueError("Function must not be empty.")

        for _ in range(self._retryTimes):
            try:
                return self._ftp_call(func, **kwargs)
            except Exception as e:
                self._logger.warning(e)
                self._logger.warning(kwargs)

            self._logger.warning("Unexpected error occurred. Retry will start in 10 sec.")
            sleep(FtpAdapter.RETRY_SEC)

        raise IOError(errno.ENOENT, "FTP failed.")

    def _ftp_call(self, func, **kwargs):
        if self._tls:
            with FTP_TLS(host=self._host, timeout=self._timeout) as ftp:  # nosec
                ftp.set_debuglevel(1)
                ftp.login(user=self._user, passwd=self._password)
                ftp.prot_p()
                return func(ftp=ftp, **kwargs)
        else:
            with FTP() as ftp:  # nosec
                ftp.set_debuglevel(1)
                ftp.connect(host=self._host, port=self._port, timeout=self._timeout)
                ftp.login(user=self._user, passwd=self._password)
                return func(ftp=ftp, **kwargs)

    def list_files(self, dir, dest, pattern):
        """
        Function that is to download files by regular pattern matching

        Args:
            dir (str): ftp server directory to download
            dest (str): local directory where to place downloaded files from ftp
            pattern (object): regular pattern matching

        Returns (tuple):
            func: list_file_func
            params: parameters for list_file_func

        Raises:
            IOError: ftplib failure
        """
        return (list_file_func, {"dir": dir, "dest": dest, "pattern": pattern})

    def clear_files(self, dir, pattern):
        """
        Remove all files which matche to pattern. Not remove directory.

        Args:
            dir (str): remove target dir
            pattern (object): remove target file pattern

        Returns (tuple):
            func: clear_file_func
            params: parameters for clear_file_func

        Raises:
            IOError: failed to remove
        """
        return (list_file_func, {"dir": dir, "pattern": pattern})

    def remove_specific_file(self, dir, fname):
        """
        Remove specific file (by name) from target directory

        Args:
            dir (str): remove target dir
            fname (str): file name (exact match)

        Returns (tuple):
            func: remove_specific_file_func
            params: parameters for remove_specific_file_func

        Raises:
            IOError: failed to remove
        """
        return (remove_specific_file_func, {"dir": dir, "fname": fname})

    def file_mdtm(self, dir, unixtime=False):
        """
        Returns dictionary of file name and timestamp from target directory

        Args:
            dir (str): directory
            unixtime=False (bool): Whether response time returns
                                   as str(yyyyMMddHHmmss) or unixtime

        Returns (tuple):
            func: file_mdtm_func
            params: parameters for file_mdtm_func

        Returns:
            dict: {file name: update time}


        Raises:
            IOError: ftplib failure
        """
        return (file_mdtm_func, {"dir": dir, "unixtime": unixtime})


def list_file_func(**kwargs):
    files = []
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        fname = os.path.basename(src)
        if kwargs["pattern"].fullmatch(fname) is None:
            continue

        os.path.join(kwargs["dir"], src)
        try:
            with open(os.path.join(kwargs["dest"], fname), "wb") as f:
                kwargs["ftp"].retrbinary("RETR " + src, f.write)
            files.append(fname)
        except IOError:
            # ignore error. try to get all files
            pass
    return files


def clear_file_func(**kwargs):
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        fname = os.path.basename(src)
        if kwargs["pattern"].fullmatch(fname) is None:
            continue
        try:
            kwargs["ftp"].delete(src)
        except Exception:
            # ignore errors. Directory cannot be deleted by ftp#delete.
            pass


def remove_specific_file_func(**kwargs):
    kwargs["ftp"].delete(os.path.join(kwargs["dir"], kwargs["fname"]))


def file_mdtm_func(**kwargs):
    res = {}
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        mdtm = kwargs["ftp"].voidcmd("MDTM %s" % src)[4:].strip()
        if kwargs["unixtime"] is True:
            mdtm = int(datetime.strptime(mdtm, "%Y%m%d%H%M%S").timestamp())
        res[os.path.basename(src)] = mdtm
    return res



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
from abc import abstractmethod
from typing import Any, List, Optional

from pydantic import BaseModel

from cliboa.adapter.file import File
from cliboa.scenario.interface import IParentStep
from cliboa.util.base import _BaseObject, _warn_deprecated


class BaseStep(_BaseObject):
    """
    Base class of all the step classes.

    ALL step classes are required to inherit this class.
    This class is minimum implemented to be needed within cliboa.
    """

    Arguments: type[BaseModel] | None = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._parent = None
        self._args = None
        self._deprecated_warn_log = []

    @property
    def logger(self) -> logging.Logger:
        """
        logger instance - set in _BaseObject.__init__
        """
        return self._logger

    @property
    def args(self) -> BaseModel | None:
        return self._args

    @property
    def parent(self) -> IParentStep | None:
        """
        parent instance which implements cliboa.scenario.interface.IParentStep
        """
        return self._parent

    @parent.setter
    def parent(self, parent: IParentStep):
        """
        set parent instance - assumed to set by cliboa
        """
        self._parent = parent

    def _set_arguments(self, arguments: dict[str, Any]) -> None:
        """
        This method allows you to set a value
        to the class with either method directly or via property setter.
        Either way, the method must be implemented
        to set the value for the class parameter like below.

        -- eg1 (Recommended) --
        from pydantic import BaseModel
        class Foo(BaseStep):
            # no need to define __init__

            class Arguments(BaseModel):
                bar: string

            def execute(self):
                # access via self.args
                self.logger.info(f"bar is {self.args.bar}")

        -- eg2 (WARNING: scheduled for deprecation in future) --
        class Foo2(BaseStep):
            def __init__(self):
                self._bar = None

            def bar(self, bar):
                self._bar = bar

        -- eg3 (WARNING: scheduled for deprecation in future) --
        class Foo3(BaseStep):
            def __init__(self):
                self._bar = None

            @property
            def bar(self):
                return self._bar

            @bar.setter
            def bar(self, bar):
                self._bar = bar
        """
        props = arguments.copy()
        if self.Arguments is not None:
            self.logger.debug(f"Set arguments using {self.__class__.__name__}.Arguments")
            self._args = self.Arguments.model_validate(props)
            for field_name, field_info in self.Arguments.model_fields.items():
                if isinstance(field_info.validation_alias, str):
                    props.pop(field_info.validation_alias, None)
                elif field_info.alias:
                    props.pop(field_info.alias, None)
                else:
                    props.pop(field_name, None)
            if len(props) > 0:
                self.logger.warning(
                    f"There are {len(props)} unused arguments after assigned to pydantic model."
                )

        self.logger.debug("Set arguments to property directly.")
        for k, v in props.items():
            if isinstance(getattr(type(self), k, None), property):
                setattr(self, k, v)
            else:
                call = getattr(self, k, None)
                if callable(call):
                    call(v)
                else:
                    if self.Arguments is not None:
                        self.logger.warning(f"Failed to set mixed argument {k}")
                    else:
                        self.logger.warning(f"Failed to set argument {k}")

    @abstractmethod
    def execute(self, *args, **kwargs) -> Optional[int]:
        """
        Main logic of step
        """
        pass

    def get_symbol_argument(self, name: str) -> Any | None:
        """
        Returns a symbol's argument (variables are already transformed).
        Returns None if the argument cannot be retrieved, and no exceptions are raised.
        """
        if not self.parent:
            return None
        sa = self.parent.get_symbol_arguments()
        return sa.get(name)

    def put_to_context(self, value: Any) -> None:
        """
        Put any given value to cliboa's context.
        The value is stored using the current step's name as the key.
        """
        if not self.parent:
            return
        self.parent.put_to_context(value)

    def get_from_context(self, target: str | None = None) -> Any | None:
        """
        Get a value from cliboa's context.
        This value is typically placed there by preceding step (the target).
        If no target is specified, the method get the value associated with the step's symbol.
        Returns None if the value is not found (not raise an exception).
        """
        if not self.parent:
            return None
        return self.parent.get_from_context(target)

    """
    Deprecated functions and properties to keep backward compatibility are below:
    """

    @property
    def _step(self) -> str:
        """
        Deprecated: Kept for v2 backward compatibility.
        """
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.base.BaseStep._step",
                "3.0",
                "4.0",
                "cliboa.scenario.base.BaseStep.put_to_context",
            )
        )
        return self.parent.step_name if self.parent else ""

    @property
    def _symbol(self) -> str | None:
        """
        Deprecated: Kept for v2 backward compatibility.
        """
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.base.BaseStep._symbol",
                "3.0",
                "4.0",
                "cliboa.scenario.base.BaseStep.get_from_context",
            )
        )
        return self.parent.symbol_name if self.parent else None

    def get_step_argument(self, name: str) -> Any | None:
        """
        Deprecated: kept for v2 backward compatibility.
        """
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.base.BaseStep.get_step_argument",
                "3.0",
                "4.0",
                "cliboa.scenario.base.BaseStep.get_symbol_argument",
            )
        )
        return self.get_symbol_argument(name)

    def get_target_files(self, src_dir: str, src_pattern: str, *args, **kwargs) -> List[str]:
        """
        Deprecated: kept for v2 backward compatibility.
        """
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.base.BaseStep.get_target_files",
                "3.0",
                "4.0",
                "cliboa.scenario.extract.file.FileRead.get_src_files",
            )
        )
        return self._resolve("adaptor_file", File).get_target_files(
            src_dir, src_pattern, *args, **kwargs
        )



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.adapter.file import File
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated_args
from cliboa.util.exception import FileNotFound


class FileRead(BaseStep):
    """
    The parent class to read the specified file
    """

    class Arguments(BaseModel):
        src_dir: str
        src_pattern: str
        encoding: str = "utf-8"
        nonfile_error: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _src_dir(self):
        return self.args.src_dir

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _src_pattern(self):
        return self.args.src_pattern

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _encoding(self):
        return self.args.encoding

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _nonfile_error(self):
        return self.args.nonfile_error

    def get_src_files(self, *args, **kwargs) -> list[str]:
        return self._resolve("adapter_file", File).get_target_files(
            self.args.src_dir, self.args.src_pattern, *args, **kwargs
        )

    def check_file_existence(self, files: list[str]) -> bool:
        """
        Check whether files exist.
        If no files, the scenario will continue or an error is raised,
        depends on parameter[nonfile_error].
        """
        if len(files) == 0:
            if self.args.nonfile_error is True:
                raise FileNotFound("No files are found.")
            else:
                self.logger.info("No files are found. Nothing to do.")
                return False
        self.logger.info("Files found %s" % files)
        return True


class FileWrite(BaseStep):
    """
    Load file to server
    """

    class Arguments(BaseModel):
        dest_dir: str | None = None
        dest_name: str | None = None
        encoding: str = "utf-8"

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dest_dir(self):
        return self.args.dest_dir

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dest_name(self):
        return self.args.dest_name



import psycopg2
from psycopg2.extras import execute_values

from cliboa.adapter.rdbms import RdbmsSupport


class PostgresqlAdaptor(RdbmsSupport):
    def get_connection(self, **kwargs):
        # see https://www.psycopg.org/docs/
        kwargs["host"] = self._host
        kwargs["user"] = self._user
        kwargs["password"] = self._password
        kwargs["database"] = self._dbname
        if self._port:
            kwargs["port"] = self._port
        return psycopg2.connect(**kwargs)

    def insert(self, sql, params):
        # psycopg2.executemany() seems like bulk insert, but it's extremely slow.
        # Instead of useing it, it changes to execute_values method, but to do so
        # the query needs to be changed like below.
        # insert into tbl_name (id, name, memo) values (%s, %s, %s)
        # ↓
        # insert into tbl_name (id, name, memo) values %s
        pos = sql.rfind("VALUES")
        sql = sql[:pos]
        sql += " VALUES %s"
        execute_values(self._con.cursor(), sql, params)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.util.base import _BaseObject, _warn_deprecated
from cliboa.util.exception import InvalidParameter

_warn_deprecated("cliboa.scenario.validator", "3.0", "4.0")


class EssentialParameters(_BaseObject):
    """
    DEPRECATED: Use cliboa.scenario.validator.EssentialParameters instead.

    Validation for the essential parameters of step class
    """

    def __init__(self, cls_name, param_list, **kwargs):
        """
        Args:
            cls_name: class name which has validation target parameters
            param_list: list of validation target parameters
        """
        super().__init__(**kwargs)
        self._cls_name = cls_name
        self._param_list = param_list
        self._logger.warning(
            _warn_deprecated(
                ".".join(("cliboa.scenario.validator", self.__class__.__name__)),
                "3.0",
                "4.0",
                "cliboa.scenario.base.BaseStep.Arguments",
            )
        )

    def __call__(self):
        for p in self._param_list:
            if not p:
                raise InvalidParameter(
                    "The essential parameter is not specified in %s." % self._cls_name
                )



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from multiprocessing import Pool

import cloudpickle
from multiprocessing_logging import install_mp_handler

from cliboa import state
from cliboa.core.interface import _IExecute
from cliboa.core.model import ParallelConfigModel
from cliboa.util.base import _BaseObject
from cliboa.util.constant import StepStatus
from cliboa.util.exception import StepExecutionFailed
from cliboa.util.log import _get_logger


class _ParallelProcessor(_BaseObject, _IExecute):
    """
    Parallel processing decorator class for _StepExecutor instances.
    """

    def __init__(
        self, steps: list[_IExecute], config: ParallelConfigModel | None = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._steps = steps
        if not config:
            config = ParallelConfigModel()
        config.fill_default()
        self._config = config

    @staticmethod
    def _async_step_execute(cls):
        try:
            clz = cloudpickle.loads(cls)
            res = clz.execute()
            if res == StepStatus.ABNORMAL_TERMINATION:
                return "NG"
            else:
                return "OK"
        except Exception as e:
            _get_logger(__name__).exception(e)
            return "NG"

    def execute(self) -> int | None:
        state.set("_ProcessParallel")
        self._logger.info(
            "Multi process start. Execute step count=%s." % self._config.multi_process_count
        )
        install_mp_handler()
        packed = [cloudpickle.dumps(step) for step in self._steps]

        try:
            with Pool(processes=self._config.multi_process_count) as p:
                for r in p.imap_unordered(self._async_step_execute, packed):
                    if r == "NG":
                        if self._config.force_continue:
                            self._logger.warning("Multi process response. %s" % r)
                        else:
                            raise StepExecutionFailed("Multi process response. %s" % r)
        except Exception:
            self._logger.exception("Exception occurred during multi process execution.")
            return StepStatus.ABNORMAL_TERMINATION



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import ABC, abstractmethod
from typing import Any


class IParentStep(ABC):
    """
    Interface for the runtime parent class of a BaseStep subclass (not the inheritance parent).
    """

    @abstractmethod
    def get_symbol_arguments(self) -> dict[str, Any]:
        """
        Returns the arguments dict for the step specified by symbol
        (variables are already transformed).
        Returns {} if symbol is not specified, and no exceptions are raised.
        """
        raise NotImplementedError()

    @abstractmethod
    def put_to_context(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get_from_context(self, step_name: str | None = None) -> Any:
        raise NotImplementedError()

    @property
    @abstractmethod
    def step_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def symbol_name(self) -> str | None:
        raise NotImplementedError()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

from pydantic import BaseModel

from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated_args


class BaseAzure(BaseStep):
    """
    Base class of Azure related classes
    """

    class Arguments(BaseModel):
        account_url: str | None = None
        account_access_key: str | None = None
        connection_string: str | None = None

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _account_url(self):
        return self.args.account_url

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _account_access_key(self):
        return self.args.account_access_key

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _connection_string(self):
        return self.args.connection_string


class BaseAzureBlob(BaseAzure):
    """
    Base class of Azure Blob Storage related classes
    """

    class Arguments(BaseAzure.Arguments):
        container_name: str

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _container_name(self):
        return self.args.container_name



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import json
import os
import re
from decimal import Decimal

import boto3

from cliboa.adapter.aws import S3Adapter
from cliboa.scenario.aws import BaseAws, BaseS3
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.constant import StepStatus
from cliboa.util.exception import InvalidParameter


class S3Download(BaseS3):
    """
    Download from S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None
        self._dest_dir = "."

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(
            self._access_key, self._secret_key, self._profile, self._role_arn, self._external_id
        )
        client = adapter.get_client()

        p = client.get_paginator("list_objects")
        keys = []
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                path = c.get("Key")
                filename = os.path.basename(path)
                rec = re.compile(self._src_pattern)
                if not rec.fullmatch(filename):
                    continue
                if self._dest_dir:
                    os.makedirs(self._dest_dir, exist_ok=True)
                dest_path = os.path.join(self._dest_dir, filename)
                client.download_file(self._bucket, path, dest_path)
                keys.append(path)

        # cache
        self.put_to_context({"bucket": self._bucket, "keys": keys})


class S3DownloadFileDelete(BaseS3):
    """
    Delete all downloaded files from S3
    """

    Arguments = None

    def execute(self, *args):
        stored = self.get_from_context()
        bucket = stored.get("bucket")
        keys = stored.get("keys")

        if keys is not None and len(keys) > 0:
            symbol_access_key = self.get_symbol_argument("access_key")
            symbol_secret_key = self.get_symbol_argument("secret_key")
            symbol_profile = self.get_symbol_argument("profile")
            symbol_role_arn = self.get_symbol_argument("role_arn")
            symbol_external_id = self.get_symbol_argument("external_id")

            adapter = S3Adapter(
                symbol_access_key,
                symbol_secret_key,
                symbol_profile,
                symbol_role_arn,
                symbol_external_id,
            )
            client = adapter.get_client()

            for key in keys:
                client.delete_object(Bucket=bucket, Key=key)
                self.logger.info("%s is successfully deleted." % bucket + "/" + key)
        else:
            self.logger.info("No files to delete.")


class S3Delete(BaseS3):
    """
    Delete from S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(
            self._access_key, self._secret_key, self._profile, self._role_arn, self._external_id
        )
        client = adapter.get_client()

        p = client.get_paginator("list_objects")
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                path = c.get("Key")
                filename = os.path.basename(path)
                rec = re.compile(self._src_pattern)
                if rec.fullmatch(filename) is None:
                    continue
                client.delete_object(Bucket=self._bucket, Key=path)


class S3FileExistsCheck(BaseS3):
    """
    File check in S3
    """

    def __init__(self):
        super().__init__()
        self._prefix = ""
        self._delimiter = ""
        self._src_pattern = None

    def prefix(self, prefix):
        self._prefix = prefix

    def delimiter(self, delimiter):
        self._delimiter = delimiter

    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        adapter = S3Adapter(
            self._access_key, self._secret_key, self._profile, self._role_arn, self._external_id
        )

        p = adapter.get_client().get_paginator("list_objects")
        for page in p.paginate(Bucket=self._bucket, Delimiter=self._delimiter, Prefix=self._prefix):
            for c in page.get("Contents", []):
                filename = os.path.basename(c.get("Key"))
                rec = re.compile(self._src_pattern)
                if not rec.fullmatch(filename):
                    continue
                # The file exist
                self._logger.info("File was found in S3. After process will be processed")
                return None

        # The file does not exist
        self._logger.info("File not found in S3. After process will not be processed")
        return StepStatus.SUCCESSFUL_TERMINATION


class DynamoDBRead(BaseAws):
    """
    Download data from DynamoDB and save as a CSV or JSONL file
    """

    def __init__(self):
        super().__init__()
        self._table_name = None
        self._dest_dir = "."
        self._file_name = None
        self._file_format = "csv"

    def table_name(self, table_name):
        self._table_name = table_name

    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def file_name(self, file_name):
        self._file_name = file_name

    def file_format(self, file_format):
        if file_format not in ["csv", "jsonl"]:
            raise InvalidParameter("file_format must be either 'csv' or 'jsonl'")
        self._file_format = file_format

    def execute(self, *args):
        """
        DynamoDBからデータをダウンロードし、指定されたフォーマットでファイルに保存します。
        """
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._table_name, self._file_name])
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        dynamodb = boto3.resource(
            "dynamodb",
            aws_access_key_id=self._access_key,
            aws_secret_access_key=self._secret_key,
            region_name=self._region,
        )
        table = dynamodb.Table(self._table_name)

        file_path = os.path.join(self._dest_dir, self._file_name)
        if self._file_format == "jsonl":
            self._write_jsonl(self._scan_table(table), file_path)
        else:
            self._write_csv(self._scan_table(table), file_path)

        self._logger.info(f"Downloaded items from DynamoDB table {self._table_name} to {file_path}")

    def _scan_table(self, table):
        """
        DynamoDBテーブルをスキャンし、全アイテムを取得するジェネレータ関数。

        Args:
            table (boto3.resources.factory.dynamodb.Table): スキャン対象のDynamoDBテーブル

        Yields:
            dict: テーブルの各アイテム
        """
        last_evaluated_key = None
        while True:
            if last_evaluated_key:
                response = table.scan(ExclusiveStartKey=last_evaluated_key)
            else:
                response = table.scan()

            for item in response["Items"]:
                yield item

            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                break

    def _write_jsonl(self, items, file_path):
        """
        アイテムをJSONL形式でファイルに書き込みます。
        Args:
            items (iterator): 書き込むアイテムのイテレータ
            file_path (str): 書き込み先のファイルパス
        """
        with open(file_path, "w") as f:
            for item in items:
                json_item = json.dumps(
                    item, default=self._json_serial, sort_keys=False, ensure_ascii=False
                )
                f.write(json_item + "\n")

    def _json_serial(self, obj):
        """
        JSONシリアライズ関数
        """
        if isinstance(obj, Decimal):
            return int(obj) if obj % 1 == 0 else float(obj)
        return str(obj)

    def _write_csv(self, items, file_path):
        """
        アイテムをCSV形式でファイルに書き込みます。

        Args:
            items (iterator): 書き込むアイテムのイテレータ
            file_path (str): 書き込み先のファイルパス
        """
        with open(file_path, "w", newline="") as f:
            writer = None
            for item in items:
                if writer is None:
                    writer = csv.DictWriter(f, fieldnames=list(item.keys()))
                    writer.writeheader()

                for key, value in item.items():
                    if isinstance(value, (dict, list)):
                        # ネストされた属性値はJSON形式に変換
                        item[key] = json.dumps(
                            value, default=self._json_serial, sort_keys=False, ensure_ascii=False
                        )

                writer.writerow(item)



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import csv
import sqlite3

from cliboa.util.base import _BaseObject


class SqliteAdapter(_BaseObject):

    _COMMIT_COUNT = 500

    """
    Adapter class of sqlite3
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._cur = None
        self._con = None

    def connect(self, dbname):
        """
        Get sqlite connection
        """
        self._con = sqlite3.connect(dbname)
        self._con.execute("PRAGMA synchronous = OFF")
        self._con.execute("PRAGMA journal_mode = OFF")

    def close(self):
        """
        Release sqlite connection
        """
        if self._cur:
            self._cur.close()
            self._cur = None

        if self._con:
            self._con.close()

    def fetch(self, sql, row_factory=None):
        """
        Get cursor after executed query of select-related

        Args:
            sql (str): SQL

        Returns:
            cursor: cursor
        """
        if self._cur:
            self._cur.close()
            self._cur = None
        self._cur = self._con.cursor()
        if row_factory:
            self._cur.row_factory = row_factory
        self._cur.execute(sql)
        return self._cur

    def create_user_func(self, dict):
        """
        Create function

        Args:
            dict: (dict)
                name: Name of user defined function.
                args: Number of arguments.
                func: User defined function.
        """
        self._con.create_function(dict["name"], dict["args"], dict["func"])

    def execute(self, sql):
        """
        Execute SQL

        Args:
            sql (str): SQL to execute
        """
        self._con.execute(sql)

    def commit(self):
        """
        Commit
        """
        self._con.commit()

    def execute_many_insert(self, tblname, column_def, insert_rows, is_replace_into=True):
        """
        Execute many INSERT INTO SQL

        Args:
            tblname: target table
            column_def: table column definition
            insert_rows(dict[])  rows to be inserted
            is_replace_into: when using replace into: True,
                             when using insert into: False
        """
        if not tblname or not insert_rows:
            raise ValueError("Parameters are missing")

        columns = ",".join(self.escape_columns(column_def))
        holders = "?" * len(column_def)
        insert_sql = "REPLACE INTO" if is_replace_into else "INSERT INTO"
        sql = insert_sql + " %s (%s) VALUES (%s)" % (
            tblname,
            columns,
            ",".join(list(holders)),
        )
        self._logger.debug("sql: %s" % sql)
        values = []
        for row in insert_rows:
            vs = []
            for c in column_def:
                vs.append(row.get(c))
            values.append(vs)

            if len(vs) != len(column_def):
                raise ValueError(
                    "The length of insert rows must be equal to the column definition. Column definition: %s, Insert rows: %s"  # noqa
                    % (column_def, vs)
                )
        self._con.executemany(sql, values)

    def create_table(self, tblname, columns, primary_key=None):
        """
        Create table, if it is not exist.

        Args:
            tblname (str): table name
            columns (str[]): column names
            primary_key=None (str) primary key
        """
        self._logger.info("Create table [%s]" % tblname)

        columns = self.escape_columns(columns)
        if primary_key is None:
            sql = "CREATE TABLE IF NOT EXISTS %s (%s)"
            self.execute(sql % (tblname, " TEXT, ".join(columns) + " TEXT"))
        else:
            sql = "CREATE TABLE IF NOT EXISTS %s (%s, PRIMARY KEY(%s))"
            self.execute(sql % (tblname, " TEXT, ".join(columns) + " TEXT", primary_key))
        self.commit()

    def drop_table(self, tblname):
        """
        Drop table, if exist.

        Args:
            tblname (str): table name
        """
        self.execute("DROP TABLE IF EXISTS %s" % tblname)
        self.commit()

    def add_columns(self, tblname, columns):
        """
        Add columns by alter table query.

        tblname (str): table name.
        columns (str[]): column names that is to be added.
        """
        for column in columns:
            self.execute("ALTER TABLE %s ADD COLUMN '%s' text" % (tblname, column))

        self.commit()

    def add_index(self, tblname, columns):
        """
        Create index

        Args:
            tblname (str): target table
            columns (str[]): index list
        """
        idx_name = tblname + "_" + "".join(columns)
        self._logger.info("Add index: %s" % idx_name)
        sql = "CREATE INDEX %s ON %s(%s)" % (idx_name, tblname, ",".join(columns))
        self.execute(sql)
        self.commit()

    def get_column_names(self, tblname):
        """
        Returns column names.

        Args:
            tblname (str): target table
        """
        cur = self.fetch("PRAGMA TABLE_INFO(%s)" % tblname)
        return [x[1] for x in cur]

    def import_table(self, src, tblname, refresh=True, encoding="utf-8", delimiter=","):
        """
        Create new table and import all data from csv(tsv).

        Args:
            src (str): csv file path
            tblname (str): table name
            refresh=True (bool): Drop table in advance, if table already exists
            encoding="utf-8" (str) Encoding
            delimiter="," (str) Set \t if src file is tsv
        """

        if refresh is True:
            self.drop_table(tblname)

        with open(src, mode="r", encoding=encoding) as f:
            reader = csv.DictReader(f, delimiter=delimiter)
            # Table columns will be the same with csv column names.

            columns = reader.fieldnames
            self.create_table(tblname, columns)

            # Put all csv records into the table.
            self._logger.info("Insert all csv records into table[%s]" % tblname)
            params = []
            count = 0
            for row in reader:
                params.append(row)
                count = count + 1
                if len(params) == self._COMMIT_COUNT:
                    self.execute_many_insert(tblname, columns, params)
                    self.commit()
                    params.clear()
            if len(params) > 0:
                self.execute_many_insert(tblname, columns, params)
                self.commit()

            self._logger.info("Insert finished. total record: %s" % count)

    def export_table(
        self,
        tblname,
        dest,
        quoting=csv.QUOTE_ALL,
        encoding="utf-8",
        order=[],
        delimiter=",",
        no_duplicate=False,
    ):
        """
        Export all data from specific table and output as csv.

        Args:
            tblname (str): table name
            dest (str): output csv file path
            encoding="utf-8" (str) Encoding
            order=[] (array) orders of outputting data
            delimiter="," (str) Set \t if src file is tsv
            no_duplicate=False (bool) If true is set, get data with "select distinct"
        """
        if no_duplicate:
            sql = "SELECT DISTINCT * FROM '%s'" % tblname
        else:
            sql = "SELECT * FROM '%s'" % tblname

        if order and len(order) > 0:
            sql += " ORDER BY " + ",".join(order)

        with open(dest, "wt", encoding=encoding, newline="") as f:
            cur = self.fetch(sql)
            fieldnames = [d[0] for d in cur.description]
            writer = csv.writer(f, quoting=quoting, delimiter=delimiter)
            writer.writerow(fieldnames)

            for row in cur:
                writer.writerow(row)
            f.flush()

    def escape_columns(self, columns):
        return ["`%s`" % column for column in columns]



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.core.context import _CliboaContext
from cliboa.core.executor import _StepExecutor
from cliboa.core.factory import _CliboaFactory, _get_scenario_loader_class
from cliboa.core.interface import _IExecute
from cliboa.core.loader import _ScenarioLoader
from cliboa.core.model import CommandArgument, ParallelStepModel, ScenarioModel, StepModel
from cliboa.core.processor import _ParallelProcessor
from cliboa.listener.base import BaseStepListener
from cliboa.listener.step import StepStatusListener
from cliboa.util.base import _BaseObject
from cliboa.util.exception import CliboaException


class _ScenarioBuilder(_BaseObject):
    """
    Build executable instances from scenario file and command argument.
    """

    def __init__(
        self,
        scenario_file: str,
        common_file: str | list[str] | None = None,
        file_format: str = "yaml",
        cmd_arg: CommandArgument | None = None,
        project_name: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._scenario_file = scenario_file
        if isinstance(common_file, str):
            self._common_files = [common_file]
        elif isinstance(common_file, list):
            self._common_files = common_file
        else:
            self._common_files = []
        self._loader_cls: _ScenarioLoader = self._resolve_cls(
            "loader", _get_scenario_loader_class(file_format)
        )
        self._scenario_model_cls = self._resolve_cls("scenario_model", ScenarioModel)
        self._factory = self._resolve("factory", _CliboaFactory, project_name)
        self._cmd_arg = cmd_arg
        self._step_model_map = {}
        self._context = self._resolve("context", _CliboaContext)

    def execute(self) -> list[_IExecute]:
        """
        Main logic of builder.
        """
        self._logger.info("Start to build scenario.")
        # 1. Get a scenario model including step models.
        scenario = self._parse_scenario()
        # 2. Set up the scenario model - propagate settings and calc with_vars.
        scenario.setup()
        # 3. Create step instances by step models which are in the scenario model.
        steps = self._create_steps(scenario)
        self._logger.info("Complete to build scenario.")
        return steps

    def _parse_scenario(self) -> ScenarioModel:
        """
        Load main scenario file and generate a model instance.

        if common file exists, merge it to main.
        """
        self._logger.info("Start to parse scenario files.")

        self._logger.info(f"Load main scenario file {self._scenario_file}")
        top_dict = self._loader_cls(self._scenario_file, True)()
        main_scenario = self._scenario_model_cls.model_validate(top_dict)

        for cmn_scenario_file in self._common_files:
            self._logger.info(f"Load common scenario file {cmn_scenario_file}")
            top_dict = self._loader_cls(cmn_scenario_file, False)()
            if top_dict:
                cmn_scenario = self._scenario_model_cls.model_validate(top_dict)
                main_scenario.merge(cmn_scenario)
            else:
                self._logger.warning(
                    f"Failed to load {cmn_scenario_file}, continue to parse scenario."
                )

        self._logger.info("Complete to parse scenario files.")
        return main_scenario

    def _setup_step_model_map(self, scenario: ScenarioModel) -> None:
        for step in scenario.scenario:
            if isinstance(step, StepModel):
                key = step.step
                if key in self._step_model_map:
                    self._logger.warning(f"Duplicate step: {key}")
                # Last-one-wins if duplicated
                self._step_model_map[key] = step
            elif isinstance(step, ParallelStepModel):
                for p_step in step.parallel:
                    key = p_step.step
                    if key in self._step_model_map:
                        self._logger.warning(f"Duplicate step: {key}")
                    # Last-one-wins if duplicated
                    self._step_model_map[key] = p_step
            else:
                raise CliboaException(f"Unexpected step instance: {step.__class__.__name__}")

    def _create_steps(self, scenario: ScenarioModel) -> list[_IExecute]:
        """
        Add executable instance to the queue
        """
        self._setup_step_model_map(scenario)
        steps = []
        for step in scenario.scenario:
            instance = self._create_step_instance(step)
            steps.append(instance)
        return steps

    def _create_step_instance(
        self, step: StepModel | ParallelStepModel
    ) -> _StepExecutor | _ParallelProcessor:
        """
        Create executable instances
        """
        if isinstance(step, StepModel):
            instance = self._create_executor(step)
            return instance
        elif isinstance(step, ParallelStepModel):
            instances = []
            for p_step in step.parallel:
                instance = self._create_executor(p_step)
                instances.append(instance)
            return self._resolve(
                "parallel_processor", _ParallelProcessor, instances, step.parallel_config
            )
        else:
            raise CliboaException(f"Unexpected step instance: {step.__class__.__name__}")

    def _create_executor(self, step: StepModel) -> _StepExecutor:
        """
        Create _StepExecutor instance - wraps BaseStep.
        """
        cls_name = step.class_name
        self._logger.debug("Create %s instance" % cls_name)

        instance = self._factory.create(cls_name, **self._di_kwargs)
        symbol_model = self._step_model_map.get(step.symbol) if step.symbol else None

        executor = self._resolve(
            "step_executor",
            _StepExecutor,
            instance,
            step,
            self._cmd_arg,
            self._context,
            symbol_model,
        )

        for lis in self._create_listeners(step):
            executor.register_listener(lis)
        return executor

    def _create_listeners(self, step: StepModel) -> list[BaseStepListener]:
        listeners = [self._resolve("step_status_listener", StepStatusListener())]
        for lis_cls in step.get_listeners():
            clz = self._factory.create(lis_cls, **self._di_kwargs)
            listeners.append(clz)
        return listeners



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
from importlib import import_module
from typing import Any

from cliboa.conf import env
from cliboa.core.loader import _JsonScenarioLoader, _ScenarioLoader, _YamlScenarioLoader
from cliboa.listener.base import BaseListener
from cliboa.scenario import *  # noqa
from cliboa.util.base import _BaseObject
from cliboa.util.exception import InvalidFormat, InvalidScenarioClass


def _get_scenario_loader_class(file_format: str) -> type[_ScenarioLoader]:
    """
    Create scenario loader instance
    """
    if file_format == "yaml":
        return _YamlScenarioLoader
    elif file_format == "json":
        return _JsonScenarioLoader
    else:
        raise InvalidFormat(f"scenario format '{file_format}' is invalid.")


class _CliboaFactory(_BaseObject):
    """
    Factory class to create cliboa step|listener instance
     - both cliboa's default class and user's custom class.
    """

    def __init__(self, project_name: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self._project_name = project_name

        self._env = self._resolve("env", env)

        self._prj_scenario_dir_name = self._env.get(
            "PROJECT_SCENARIO_DIR_NAME", self._env.get("SCENARIO_DIR_NAME")
        )
        self._cmn_classes = self._env.COMMON_CUSTOM_CLASSES
        self._prj_classes = self._env.PROJECT_CUSTOM_CLASSES
        self._cmn_root_paths = self._env.get("COMMON_CUSTOM_ROOT_PATHS")
        if self._cmn_root_paths is None:
            self._cmn_root_paths = [self._restore_v2_cmn_root_path()]
        self._prj_root_paths = self._env.get("PROJECT_CUSTOM_ROOT_PATHS")
        if self._prj_root_paths is None:
            self._prj_root_paths = [self._restore_v2_prj_root_path()]

    def _restore_v2_cmn_root_path(self) -> str:
        """
        Restore v2 common root path for backward compability.
        """
        base_dir = self._env.get("BASE_DIR")
        common_scenario_dir_name = self._env.get("COMMON_SCENARIO_DIR")
        if (
            base_dir is not None
            and common_scenario_dir_name is not None
            and common_scenario_dir_name.startswith(base_dir)
        ):
            common_scenario_root_dir = common_scenario_dir_name[len(base_dir) + len(os.sep) :]
            return ".".join(common_scenario_root_dir.split(os.sep))
        else:
            return "common.scenario"

    def _restore_v2_prj_root_path(self) -> str:
        """
        Restore v2 project root path for backward compability.
        """
        base_dir = self._env.get("BASE_DIR")
        prj_dir_name = self._env.get("PROJECT_DIR")
        if base_dir is not None and prj_dir_name is not None and prj_dir_name.startswith(base_dir):
            prj_root_dir = prj_dir_name[len(base_dir) + len(os.sep) :]
            return ".".join(prj_root_dir.split(os.sep))
        else:
            return "project"

    def _import_custom_cls(
        self,
        mod_name: str,
        sliced_paths: list[str],
        root_paths: list[str],
        is_prj: bool,
    ) -> type[Any]:
        errors: list[str] = []
        for root_path in root_paths:
            try:
                if root_path == "":
                    import_root_paths = []
                else:
                    import_root_paths = root_path.split(".")
                if is_prj:
                    import_root_paths.append(self._project_name)
                    if self._prj_scenario_dir_name is not None:
                        import_root_paths.append(self._prj_scenario_dir_name)
                import_root_paths += sliced_paths
                root = ".".join(import_root_paths)
                return getattr(import_module(root), mod_name)
            except (ImportError, AttributeError) as e:
                errors.append(str(e))
        cls_name = ".".join(sliced_paths) + "." + mod_name
        raise ImportError(f"Failed to cliboa import {cls_name}:\n" + "\n".join(errors))

    def _get_custom_cls(
        self, cls_name: str, root_paths: list[str], custom_classes: list[str], is_prj: bool = False
    ) -> type[Any] | None:
        for custom_class_path in custom_classes:
            sliced_paths = custom_class_path.split(".")
            mod_name = sliced_paths.pop()
            if mod_name != cls_name:
                continue
            return self._import_custom_cls(mod_name, sliced_paths, root_paths, is_prj)
        return None

    def create(self, cls_name: str, **kwargs) -> object:
        """
        Create cliboa step|listener instance.
        """
        try:
            cls = self._get_custom_cls(cls_name, self._cmn_root_paths, self._cmn_classes)
            if cls is None and self._project_name:
                cls = self._get_custom_cls(
                    cls_name, self._prj_root_paths, self._prj_classes, is_prj=True
                )
            if cls is None:
                cls = globals()[cls_name]
            instance = cls(**kwargs)
            if not isinstance(
                instance,
                (
                    BaseStep,  # noqa
                    BaseListener,
                ),
            ):
                self._logger.warning(
                    f"{instance.__class__.__name__} does not inherit from "
                    "cliboa.scenario.base.BaseStep or cliboa.listener.base.BaseListener"
                )
            return instance
        except Exception:
            self._logger.exception(f"Failed to create an instance of {cls_name}")
            raise InvalidScenarioClass(f"Invalid scenario class '{cls_name}'")



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import logging
import os
import sys
from typing import List

from cliboa import state


def _get_logger(modname: str):
    """
    Get logger

    Args:
        modname (str): module name
    Returns:
        logger instance
    """
    return logging.getLogger(modname)


_sys_uniq_path = []
_last_sys_path_count = -1


def _generate_sys_uniq_path() -> List[str]:
    seen = set()
    abs_paths = []
    for path in sys.path:
        if not path:
            continue
        abs_path = os.path.abspath(path)
        if abs_path not in seen:
            seen.add(abs_path)
            abs_paths.append(abs_path)

    paths_to_remove = set()

    for path1 in abs_paths:
        if path1 in paths_to_remove:
            continue
        for path2 in abs_paths:
            if path1 == path2 or path2 in paths_to_remove:
                continue

            longer_path, shorter_path = (None, None)

            if path1.startswith(path2) and len(path1) > len(path2):
                longer_path, shorter_path = path1, path2
            elif path2.startswith(path1) and len(path2) > len(path1):
                longer_path, shorter_path = path2, path1
            else:
                continue

            if longer_path.endswith("site-packages"):
                continue

            diff_path = os.path.relpath(longer_path, shorter_path)
            diff_parts = diff_path.split(os.sep)
            if any(part.startswith(".") for part in diff_parts if part):
                continue

            paths_to_remove.add(longer_path)

    uniq_paths = [p for p in abs_paths if p not in paths_to_remove]
    return uniq_paths


def _create_module_path(record_path) -> str:
    global _sys_uniq_path, _last_sys_path_count
    current_sys_path_count = len(sys.path)
    if current_sys_path_count != _last_sys_path_count:
        _sys_uniq_path = _generate_sys_uniq_path()
        _last_sys_path_count = current_sys_path_count

    rel_path = None
    for sys_path in _sys_uniq_path:
        if record_path.startswith(os.path.join(sys_path, "")):
            rel_path = record_path[len(sys_path) :].lstrip(os.sep)
            break

    if not rel_path:
        rel_path = record_path.lstrip(os.sep)

    module_path, _ = os.path.splitext(rel_path)
    parts = module_path.split(os.sep)
    if len(parts) >= 2 and parts[-1] == "__init__":
        parts.pop()

    s_parts = []
    full_flg = False
    last_i = len(parts) - 1
    for i, p in enumerate(parts):
        if full_flg:
            s_parts.append(p)
            full_flg = False
        elif p == "project" and i < last_i:
            s_parts.append(p[0])
            full_flg = True
        elif i == 0 and p == "common":
            s_parts.append(p[0])
        elif i == 0 or i == last_i:
            s_parts.append(p)
        else:
            s_parts.append(p[0])
    return ".".join(s_parts)


class CliboaLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # default values
        self.cliboaState = "_Unknown"
        self.cliboaPath = self.module
        # set values
        try:
            if not self.processName or self.processName == "MainProcess":
                self.cliboaState = str(state)
            else:
                self.cliboaState = self.processName + "." + str(state)
            self.cliboaPath = _create_module_path(self.pathname)
        except Exception:
            pass



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from pydantic import Field

from cliboa.scenario.sqlite import BaseSqlite
from cliboa.util.base import _warn_deprecated_args


class SqliteExport(BaseSqlite):
    class Arguments(BaseSqlite.Arguments):
        tblname: str
        dest_path: str
        encoding: str = "utf-8"
        order: list[str] = Field(default_factory=list)
        no_duplicate: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _tblname(self):
        return self.args.tblname

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dest_path(self):
        return self.args.dest_path

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _encoding(self):
        return self.args.encoding

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _order(self):
        return self.args.order

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _no_duplicate(self):
        return self.args.no_duplicate

    def execute(self, *args):
        dest_dir = os.path.dirname(self.args.dest_path)
        if dest_dir:
            os.makedirs(dest_dir, exist_ok=True)

        self._sqlite_adptr.connect(self.args.dbname)
        try:
            self._sqlite_adptr.export_table(
                self.args.tblname,
                self.args.dest_path,
                encoding=self.args.encoding,
                order=self.args.order,
                no_duplicate=self.args.no_duplicate,
            )
        finally:
            self._close_database()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.adapter.ftp import FtpAdapter
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated


class BaseFtp(BaseStep):
    class Arguments(BaseModel):
        src_dir: str
        src_pattern: str
        dest_dir: str = ""
        host: str
        port: int = 21
        user: str
        password: str | None = None
        timeout: int = 30
        retry_count: int = 3
        tls: bool = False

    def get_adapter(self):
        return self._resolve(
            "adapter_ftp",
            FtpAdapter,
            host=self.args.host,
            user=self.args.user,
            password=self.args.password,
            timeout=self.args.timeout,
            retryTimes=self.args.retry_count,
            port=self.args.port,
            tls=self.args.tls,
        )

    def get_adaptor(self):
        self.logger.info(
            _warn_deprecated(
                "cliboa.scenario.ftp.BaseFtp.get_adaptor",
                "3.0",
                "4.0",
                "cliboa.scenario.ftp.BaseFtp.get_adapter",
            )
        )
        return self.get_adapter()



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
"""
Exception classes of cliboa
"""


class CliboaException(Exception):
    # cliboa unique exception
    pass


class InvalidScenarioClass(CliboaException):
    # Exception when scenario class is invalid
    pass


class DirStructureInvalid(CliboaException):
    # Exception when directory structure invalid
    pass


class FileNotFound(CliboaException):
    # Exception when specified file not found
    pass


class InvalidFileCount(CliboaException):
    # Exception when many files exist more than expacted
    pass


class InvalidParameter(CliboaException):
    # Exception when parameters are invalid
    pass


class InvalidCount(CliboaException):
    # Exception when invalid count was specified
    pass


class InvalidFormat(CliboaException):
    # Exception when invalid format was specified
    pass


class ScenarioFileInvalid(CliboaException):
    # Exception when scenario.yml(json) is invalid
    pass


class StepExecutionFailed(CliboaException):
    # Exception when step execution was failed
    pass


class DatabaseException(CliboaException):
    # Database error
    pass



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.file import FileRead as _FR
from cliboa.util.base import _warn_deprecated

_warn_deprecated(
    "cliboa.scenario.load.file",
    "3.0",
    "4.0",
    "cliboa.scenario.file",
)


class FileRead(_FR):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger.warning(
            _warn_deprecated(
                "cliboa.scenario.load.file.FileRead",
                "3.0",
                "4.0",
                "cliboa.scenario.file.FileRead",
            )
        )



# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import boto3
from boto3.session import Session

from cliboa.util.base import _BaseObject
from cliboa.util.exception import InvalidParameter


class S3Adapter(_BaseObject):
    """
    Adapter of AWS S3
    """

    def __init__(
        self,
        access_key: str = None,
        secret_key: str = None,
        profile: str = None,
        role_arn: str = None,
        external_id: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Validate credential combinations
        # 1) access_key and secret_key must be specified together
        if (access_key is None) ^ (secret_key is None):
            raise InvalidParameter("access_key and secret_key must be specified together.")

        # 2) Allow at most one of {profile, access_key/secret_key, role_arn}
        auth_methods = [bool(profile), bool(access_key and secret_key), bool(role_arn)]
        if sum(auth_methods) > 1:
            raise InvalidParameter(
                "Specify only one of profile, access_key/secret_key, or role_arn."
            )

        self._access_key = access_key
        self._secret_key = secret_key
        self._profile = profile
        self._role_arn = role_arn
        self._external_id = external_id

    def get_client(self):
        """
        Get s3 client
        """
        if self._profile:
            return Session(profile_name=self._profile).client("s3")
        elif self._access_key and self._secret_key:
            return boto3.client(
                "s3",
                aws_access_key_id=self._access_key,
                aws_secret_access_key=self._secret_key,
            )
        elif self._role_arn:
            return self._get_cross_account_client()
        else:
            return boto3.client("s3")

    def get_resource(self):
        """
        Get s3 resource
        """
        if self._profile:
            return Session(profile_name=self._profile).resource("s3")
        elif self._access_key and self._secret_key:
            session = Session(
                aws_access_key_id=self._access_key,
                aws_secret_access_key=self._secret_key,
            )
            return session.resource("s3")
        elif self._role_arn:
            return self._get_cross_account_resource()
        else:
            return boto3.resource("s3")

    def _get_cross_account_client(self):
        """
        Get S3 client using cross-account IAM role
        """
        credentials = self._assume_role_credentials()

        return boto3.client(
            "s3",
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )

    def _get_cross_account_resource(self):
        """
        Get S3 resource using cross-account IAM role
        """
        credentials = self._assume_role_credentials()

        session = Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )

        return session.resource("s3")

    def _assume_role_credentials(self):
        """Assume the configured role and return temporary credentials dict."""
        sts_client = boto3.client("sts")

        assume_role_kwargs = {"RoleArn": self._role_arn, "RoleSessionName": "cliboa-session"}

        if self._external_id:
            assume_role_kwargs["ExternalId"] = self._external_id

        try:
            response = sts_client.assume_role(**assume_role_kwargs)
        except Exception as e:
            raise InvalidParameter(f"Failed to assume role {self._role_arn}: {str(e)}")

        return response["Credentials"]



#
# Copyright BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from pydantic import BaseModel

from cliboa.adapter.sqlite import SqliteAdapter
from cliboa.scenario.base import BaseStep
from cliboa.util.base import _warn_deprecated_args
from cliboa.util.constant import StepStatus


class BaseSqlite(BaseStep):
    """
    Base class of all the sqlite classes
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._sqlite_adptr = self._resolve("adapter_sqlite", SqliteAdapter)

    class Arguments(BaseModel):
        dbname: str
        vacuum: bool = False

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _dbname(self):
        return self.args.dbname

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def _vacuum(self):
        return self.args.vacuum

    def _dict_factory(self, cursor, row):
        d = {}
        for i, col in enumerate(cursor.description):
            d[col[0]] = row[i]
        return d

    def _close_database(self):
        """
        Disconnect sqlite database (execute vacuum if necessary)
        """
        self._sqlite_adptr.close()
        if self.args.vacuum is True:
            try:
                self._sqlite_adptr.connect(self.args.dbname)
                self._sqlite_adptr.execute("VACUUM")
            finally:
                self._sqlite_adptr.close()


class SqliteTransaction(BaseSqlite):
    """
    Base class of sqlite transaction
    """

    def execute(self, *args):
        """
        Execute db connection, commit, close, vacuum

        Args:
            args[0]: db operation function to execute as transaction
        """

        # TODO
        # Add to handle to select DB transaction seperation level
        self.logger.info("Start DB Transaction")
        self._sqlite_adptr.connect(self.args.dbname)
        try:
            self.process()
            self._sqlite_adptr.commit()
        except NotImplementedError:
            return StepStatus.ABNORMAL_TERMINATION
        finally:
            super()._close_database()
        self.logger.info("Finish DB Transaction")

    def process(self):
        raise NotImplementedError()


class SqliteQueryExecute(SqliteTransaction):
    """
    Execute only row query of insert, update or delete
    """

    class Arguments(SqliteTransaction.Arguments):
        raw_query: str

    @property
    @_warn_deprecated_args("3.0", "4.0")
    def raw_query(self, raw_query):
        self.args.raw_query = raw_query

    def process(self):
        self._sqlite_adptr.execute(self.args.raw_query)
